# -*- coding: utf-8 -*-
"""
Created on Sun Nov  8 10:10:29 2020

@author: rusli001
"""

# import necessary package
import os
import flopy
import flopy.utils.binaryfile as bf
import flopy.discretization as fgrid
from flopy.utils.gridintersect import GridIntersect
import numpy as np
import math
import xarray as xr
import geopandas as gpd
import pandas as pd
from scipy.interpolate import griddata
from scipy.interpolate import Rbf
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

# create the model and define the workspace
modelname = 'gwf_model'
exe_name = 'exe/mf6.exe'
model_ws = 'E:/test'

# create the basic object: simulation, model solution, and groundwater flow model
sim = flopy.mf6.MFSimulation(sim_name = modelname, exe_name = exe_name, sim_ws = model_ws)
ims = flopy.mf6.ModflowIms(sim, print_option = "SUMMARY", complexity = "SIMPLE", outer_dvclose = 0.01, outer_maximum = 100,
                           under_relaxation = 'NONE', inner_maximum = 100, inner_dvclose = 0.01, linear_acceleration = 'CG',
                           preconditioner_levels = 7, preconditioner_drop_tolerance = 0.01, number_orthogonalizations = 2)
sim.register_ims_package(ims, [modelname])
gwf = flopy.mf6.ModflowGwf(sim, modelname = modelname, save_flows = True)

# create the time discretization: 14 years period with monthly timestep and daily sub-timestep
nper = (2018-2005+1)*12
period_data = [(30.0, 120, 1.0)] * nper
tdis = flopy.mf6.ModflowTdis(sim, time_units = "DAYS", nper = 168, perioddata = period_data)

################################# DIS package #######################################################################################

# get shapefile
basin = gpd.read_file('shp/groundwater basin/domain_clip4.shp')

# define boundary
basin_geometry = basin.iloc[0].geometry
basin_width = basin_geometry.bounds[2] - basin_geometry.bounds[0]
basin_height = basin_geometry.bounds[3] - basin_geometry.bounds[1]

# define spatial discretization with dummy elevation (DIS_0.1 package)
top = 100
botm = 0
nlay = 4
delr = 100
delc = 100
nrow = math.ceil(basin_height / delc)
ncol = math.ceil(basin_width / delr)

s_grid = fgrid.StructuredGrid(delc * np.ones(nrow), delr * np.ones(ncol), xoff = basin_geometry.bounds[0], yoff = basin_geometry.bounds[1])
ix = GridIntersect(s_grid, method = 'vertex', rtree = True)

basin_intersect = ix.intersect(basin_geometry)

idomain = np.zeros((nlay, nrow, ncol))
for i in basin_intersect.cellids:
    idomain[:, i[0], i[1]] = 1
for i in basin_intersect.cellids:
    idomain[3, i[0], i[1]] = 0

# define initial (bound to change) DIS_0 package
dis = flopy.mf6.ModflowGwfdis(gwf, length_units = "METERS", nlay = nlay, ncol = ncol, nrow = nrow,
                              delr = delr, delc = delc, top = top, botm = botm, idomain = idomain)

gwf.modelgrid.set_coord_info(xoff = basin_geometry.bounds[0], yoff = basin_geometry.bounds[1], angrot = 0, epsg = 32748)

# assign MERIT-DEM-based elevation
dem = pd.read_csv('csv/DIS - surface - MERIT-DEM-based elevation.csv')
points = dem[['Easting', 'Northing']].values
values = dem['Elevation'].values
grid_x = gwf.modelgrid.xcellcenters
grid_y = gwf.modelgrid.ycellcenters
dis_top = griddata(points, values, (grid_x, grid_y), method = 'nearest')

# update DIS_0.1 package with MERIT-DEM surface elevation
gwf.dis.top = dis_top

# define subsurface elevation with dummy values (DIS_0.2 package)
bottom_elev = 100
dummy_botm = np.zeros((nlay, nrow, ncol))
dummy_botm[0,:,:] = dis_top - 100
dummy_botm[1,:,:] = dummy_botm[0,:,:] - 100
dummy_botm[2,:,:] = bottom_elev + 100
dummy_botm[3,:,:] = bottom_elev

# assign the first layer depth (upper aquifer bottom / aquitard top)
dis_lay0_botm = pd.read_csv('csv/DIS - layer 1 bottom.csv')
points_lay0 = dis_lay0_botm[['X', 'Y']].values
depth_lay0 = dis_lay0_botm['Depth'].values
grid_x = gwf.modelgrid.xcellcenters
grid_y = gwf.modelgrid.ycellcenters
dis_lay0_depth = griddata(points_lay0, depth_lay0, (grid_x, grid_y), method = 'nearest')

dummy_botm[0,:,:] = dis_top - dis_lay0_depth

# adjust the first layer depth (upper aquifer bottom / aquitard top)
dummy_botm[0,:,:] = np.nanmin(dummy_botm[0,:,:]) + 0.25*(dis_top - np.nanmin(dummy_botm[0,:,:]))

# assign the second layer depth (aquitard bottom / lower aquifer top)
dis_lay1_botm = pd.read_csv('csv/DIS - layer 2 bottom.csv')
points_lay1 = dis_lay1_botm[['X', 'Y']].values
depth_lay1 = dis_lay1_botm['Relative depth'].values
grid_x = gwf.modelgrid.xcellcenters
grid_y = gwf.modelgrid.ycellcenters
dis_lay1_depth = griddata(points_lay1, depth_lay1, (grid_x, grid_y), method = 'nearest')

dummy_botm[1,:,:] = dummy_botm[0,:,:] - dis_lay1_depth

# update DIS_0.1 package with dummy subsurface layers
gwf.dis.botm = dummy_botm

#####################################################################################################################################

################################## IC package #######################################################################################

# read previous run hds file
hobj = flopy.utils.HeadFile('E:/PhD/05 Analysis/003 Paper 3/006 20220221 - final model modified/simulation results/3 - full abstraction/gwf_model.hds')
#hobj = flopy.utils.HeadFile('steady-state_result/gwf_model.hds')
ic_1 = hobj.get_data(totim = 5040.0)
#ic_1_ = ic_1[:, :, 15:646]

# define the initial conditions
ic = flopy.mf6.ModflowGwfic(gwf, strt = ic_1)

#####################################################################################################################################

################################# NPF package #######################################################################################

# read shallow groundwater hydraulic conductivity from field data
k_1 = pd.read_csv('csv/NPF - layer 1 - field data - edit.csv')
k_1_points = k_1[['X (actual)', 'Y (actual)']].values
k_1_values = k_1['k'].values
k33_1_values = k_1['k33'].values

# read deep groundwater hydraulic conductivity from pumping tests
k_3 = pd.read_csv('csv/NPF - layer 3 - field data - edit.csv')
k_3_points = k_3[['X (actual)', 'Y (actual)']].values
k_3_values = k_3['k'].values

# interpolate using radial basis function from scipy.interpolate
rbf_npf_k1 = Rbf(k_1_points[:,0], k_1_points[:,1], k_1_values[:], function = "thin_plate")
grid_k_1_spline = rbf_npf_k1(grid_x, grid_y)

rbf_npf_k3 = Rbf(k_3_points[:,0], k_3_points[:,1], k_3_values[:], function = "thin_plate")
grid_k_3_spline = rbf_npf_k3(grid_x, grid_y)

rbf_npf_k33 = Rbf(k_1_points[:,0], k_1_points[:,1], k33_1_values[:], function = "thin_plate")
grid_k_33_spline = rbf_npf_k33(grid_x, grid_y)

k_factor = 1
aquitard_factor = 0.11

# create hydraulic conductivity array
flopy_npf_k = np.zeros((nlay, nrow, ncol))
flopy_npf_k[0,:,:] = k_factor*grid_k_1_spline
flopy_npf_k[1,:,:] = aquitard_factor*grid_k_1_spline
flopy_npf_k[2,:,:] = grid_k_3_spline*0.8
flopy_npf_k[3,:,:] = 0.0000001

flopy_npf_k33 = np.zeros((nlay, nrow, ncol))
flopy_npf_k33[0,:,:] = k_factor*grid_k_33_spline*10
flopy_npf_k33[1,:,:] = aquitard_factor*grid_k_33_spline*10
flopy_npf_k33[2,:,:] = grid_k_3_spline*1
flopy_npf_k33[3,:,:] = 0.0000001

# replace negative value in the interpolation to minimum value (applicable only to RBF interpolation)
flopy_npf_k[0,:,:][flopy_npf_k[0,:,:] < 0.0003504253562667259] = 0.0003504253562667259
flopy_npf_k[1,:,:][flopy_npf_k[1,:,:] < 3.504253562667259e-06] = 3.504253562667259e-06
flopy_npf_k[2,:,:][flopy_npf_k[2,:,:] < 0.0003504253562667259] = 0.0003504253562667259
flopy_npf_k33[0,:,:][flopy_npf_k33[0,:,:] < 5.987984615790642e-08] = 5.987984615790642e-08
flopy_npf_k33[1,:,:][flopy_npf_k33[1,:,:] < 5.987984615790642e-10] = 5.987984615790642e-10
flopy_npf_k33[2,:,:][flopy_npf_k33[2,:,:] < 5.987984615790642e-08] = 5.987984615790642e-08

# assign cell type
icelltype = np.zeros((nlay, nrow, ncol))
for i in basin_intersect.cellids:
    icelltype[0, i[0], i[1]] = 1
    icelltype[1, i[0], i[1]] = 0
    icelltype[2, i[0], i[1]] = 0
    icelltype[3, i[0], i[1]] = 0

# define node property flow (NPF) package, icelltype = 0 = confined
npf = flopy.mf6.ModflowGwfnpf(gwf, save_specific_discharge = True, icelltype = icelltype, k = flopy_npf_k, k33 = flopy_npf_k33,
                             rewet_record = ['wetfct', 1.0, 'iwetit', 1, 'ihdwet', 0], wetdry = 1.)

#####################################################################################################################################

################################# RCH package #######################################################################################

# define recharge file source name
date_to_num = pd.read_excel('Date_to_num.xlsx', sheet_name = 'month')
month_year_name = date_to_num['Name'].values

# read loopified recharge file
rch_read = []

i = 0
for i in range(nper):
    rch_read.append(xr.open_dataset('netcdf/maxleakage=0.9_clip4/recharge_%s.nc' % month_year_name[i+1]))
    i += 1

# prepare the recharge file to readable input to MODFLOW
rch_read_ = []

i = 0
for i in range(nper):
    rch_read[i] = ((rch_read[i].fillna(0.9)).to_array(name = 'recharge')[1])/1000
    rch_read_.append(rch_read[i][0])
    i += 1

rch_list = [[] for i in range(nper)]

for i in range(nper):
    for j in range(len(basin_intersect)):
        rch_list[i].append([(0, basin_intersect[j][0][0], basin_intersect[j][0][1]), rch_read_[i].values[basin_intersect[j][0][0], basin_intersect[j][0][1]]])

# define loopified rch_list read to the final rch_input
rch_input = {}

i = 0
for i in range(nper):
    rch_input[i] = rch_list[i][:]
    i += 1

rch = flopy.mf6.ModflowGwfrch(gwf, stress_period_data = rch_input)

#####################################################################################################################################

################################# STO package #######################################################################################

# assign iconvert type
iconvert = np.zeros((nlay, nrow, ncol))
for i in basin_intersect.cellids:
    iconvert[0, i[0], i[1]] = 1
    iconvert[1, i[0], i[1]] = 0
    iconvert[2, i[0], i[1]] = 0
    iconvert[3, i[0], i[1]] = 0

# assign specific storage values
ss = np.zeros((nlay, nrow, ncol))
for i in basin_intersect.cellids:
    ss[:, i[0], i[1]] = 0.2
for i in basin_intersect.cellids:
    ss[1, i[0], i[1]] = 8.7e-3
for i in basin_intersect.cellids:
    ss[2, i[0], i[1]] = 8.7e-3

transient = {}

i = 0
for i in range(nper):
    transient[i] = True
    i += 1

#define dynamic hydraulic parameters: iconvert = 0 = confined
sto = flopy.mf6.ModflowGwfsto(gwf, save_flows = True, iconvert = iconvert, ss = ss, sy = 0.2, transient = transient)

#####################################################################################################################################

################################# RIV package #######################################################################################

# get shapefile boundary & create the 'cell-line'
river = gpd.read_file('shp/river path/riv_fac100.shp')
river_geometry = river.iloc[0].geometry
river_intersect = ix.intersect(river_geometry)

river_list = []
for i in river_intersect.cellids:
    river_list.append([(0, i[0], i[1]), dis_top[i[0], i[1]] - 21, 75, dis_top[i[0], i[1]] - 23.5])

river_spd = {}

i = 0
for i in range(nper):
    river_spd[i] = river_list
    i += 1

# define the RIV package
riv = flopy.mf6.ModflowGwfriv(gwf, stress_period_data = river_spd)

#####################################################################################################################################

################################# WEL package #######################################################################################

# get point location shapefile & its cellids for domestic and industrial abstraction
abs_dom_well = gpd.read_file('shp/abstraction point/domestic_new4.shp')
abs_dom_well_geometry = []

i = 0
for i in range(len(abs_dom_well)):
    abs_dom_well_geometry.append(ix.intersect(abs_dom_well.iloc[i].geometry))
    i += 1

abs_ind_well = gpd.read_file('shp/abstraction point/industry_point-river.shp')
abs_ind_well_geometry = []

i = 0
for i in range(len(abs_ind_well)):
    abs_ind_well_geometry.append(ix.intersect(abs_ind_well.iloc[i].geometry))
    i += 1

# read the abstraction volume rate estimate and prepare the wel rate input
abs = pd.read_csv('csv/WEL - abstraction_2005-2018.csv')
abs_dom = abs['Domestic groundwater abstraction'].values[1:].astype(float)
ind_dom = abs['Industrial groundwater abstraction'].values[1:].astype(float)

# assign the average domestic and industry abstraction to the wel stress period data
abs_well = []

i = 0
for i in range(len(abs_dom)):
    abs_well.append(flopy.mf6.ModflowGwfwel.stress_period_data.empty(gwf, maxbound = len(abs_dom_well_geometry) + len(abs_ind_well_geometry)))
    i += 1

i = 0
j = 0
for i in range(len(abs_dom)):
    for j in range(len(abs_dom_well)):
        abs_well[i][0][j] = ((0, abs_dom_well_geometry[j][0][0][0], abs_dom_well_geometry[j][0][0][1]), -1*abs_dom[i]/len(abs_dom_well_geometry))
        j += 1
    i += 1

i = 0
j = len (abs_dom_well_geometry)
for i in range(len(abs_dom)):
    for j in range(len(abs_dom_well_geometry), len(abs_dom_well_geometry) + len(abs_ind_well_geometry), 1):
        abs_well[i][0][j] = ((2, abs_ind_well_geometry[j - len(abs_dom_well_geometry)][0][0][0], abs_ind_well_geometry[j - len(abs_dom_well_geometry)][0][0][1]), -1*ind_dom[i]/len(abs_ind_well_geometry))
        j += 1
    i += 1

# define loopified abs_wel read to the final wel_input
wel_input = {}
wel_input_key = {}

i = 0
for i in range(len(abs_well)):
    wel_input[i] = abs_well[i][0]
    wel_input_key[i] = i*12
    i += 1

wel_input = dict((wel_input_key[key], value) for (key, value) in wel_input.items())

wel = flopy.mf6.ModflowGwfwel(gwf, stress_period_data = wel_input, save_flows = True)

#####################################################################################################################################

################################## OC package #######################################################################################

hname = '{}.hds'.format(modelname)
cname = '{}.cbc'.format(modelname)
oc = flopy.mf6.ModflowGwfoc(gwf, head_filerecord = hname, budget_filerecord= cname, saverecord = [('HEAD', 'ALL'), ('BUDGET', 'ALL')])

#####################################################################################################################################

################################## MODEL RUN ########################################################################################

# write the model
sim.write_simulation()
#sim.run_simulation()

#####################################################################################################################################