# -*- coding: utf-8 -*-
"""
Created on Sun Nov  8 10:10:29 2020

@author: rusli001
"""

# import necessary package
import os
import flopy
import flopy.utils.binaryfile as bf
import flopy.discretization as fgrid
from flopy.utils.gridintersect import GridIntersect
import numpy as np
import math
import xarray as xr
import geopandas as gpd
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
import pandas as pd
from scipy.interpolate import griddata
from scipy.interpolate import Rbf

# create the model and define the workspace
modelname = 'gwf_model'
exe_name = 'exe/mf6.exe'
model_ws = 'test'

# create the basic object: simulation, model solution, and groundwater flow
sim = flopy.mf6.MFSimulation(sim_name = modelname, exe_name = exe_name, sim_ws = model_ws)
ims = flopy.mf6.ModflowIms(sim, print_option = "SUMMARY", complexity = "SIMPLE", outer_dvclose = 0.01, outer_maximum = 100,
                           under_relaxation = 'NONE', inner_maximum = 100, inner_dvclose = 0.01, linear_acceleration = 'CG',
                           preconditioner_levels = 7, preconditioner_drop_tolerance = 0.01, number_orthogonalizations = 2)
sim.register_ims_package(ims, [modelname])
gwf = flopy.mf6.ModflowGwf(sim, modelname = modelname, save_flows = True)

# create the time discretization: 50 years of steady state and 4 years with constant abstraction
tdis = flopy.mf6.ModflowTdis(sim, time_units = "DAYS", nper = 1, perioddata = [(1.0, 1, 1.0)])

################################# DIS package #######################################################################################

# get shapefile
basin = gpd.read_file('shp/groundwater basin/domain_clip4.shp')

# define boundary
basin_geometry = basin.iloc[0].geometry
basin_width = basin_geometry.bounds[2] - basin_geometry.bounds[0]
basin_height = basin_geometry.bounds[3] - basin_geometry.bounds[1]

# define spatial discretization with dummy elevation (DIS_0.1 package)
top = 100
botm = 0
nlay = 4
delr = 100
delc = 100
nrow = math.ceil(basin_height / delc)
ncol = math.ceil(basin_width / delr)

s_grid = fgrid.StructuredGrid(delc * np.ones(nrow), delr * np.ones(ncol), xoff = basin_geometry.bounds[0], yoff = basin_geometry.bounds[1])

ix = GridIntersect(s_grid, method = 'vertex', rtree = True)

basin_intersect = ix.intersects(basin_geometry)

idomain = np.zeros((nlay, nrow, ncol))
for i in basin_intersect.cellids:
    idomain[:, i[0], i[1]] = 1
for i in basin_intersect.cellids:
    idomain[3, i[0], i[1]] = 0

# define initial (bound to change) DIS_0 package
dis = flopy.mf6.ModflowGwfdis(gwf, length_units = "METERS", nlay = nlay, ncol = ncol, nrow = nrow,
                              delr = delr, delc = delc, top = top, botm = botm, idomain = idomain)

gwf.modelgrid.set_coord_info(xoff = basin_geometry.bounds[0], yoff = basin_geometry.bounds[1], angrot = 0, epsg = 32748)

# assign MERIT-DEM-based elevation
dem = pd.read_csv('csv/DIS - surface - MERIT-DEM-based elevation.csv')
points = dem[['Easting', 'Northing']].values
values = dem['Elevation'].values
grid_x = gwf.modelgrid.xcellcenters
grid_y = gwf.modelgrid.ycellcenters
dis_top = griddata(points, values, (grid_x, grid_y), method = 'nearest')

# update DIS_0.1 package with MERIT-DEM surface elevation
gwf.dis.top = dis_top

# define subsurface elevation with dummy values (DIS_0.2 package)
bottom_elev = 100
dummy_botm = np.zeros((nlay, nrow, ncol))
dummy_botm[0,:,:] = dis_top - 100
dummy_botm[1,:,:] = dummy_botm[0,:,:] - 100
dummy_botm[2,:,:] = bottom_elev + 100
dummy_botm[3,:,:] = bottom_elev

# assign the first layer depth (upper aquifer bottom / aquitard top)
dis_lay0_botm = pd.read_csv('csv/DIS - layer 1 bottom.csv')
points_lay0 = dis_lay0_botm[['X', 'Y']].values
depth_lay0 = dis_lay0_botm['Depth'].values
grid_x = gwf.modelgrid.xcellcenters
grid_y = gwf.modelgrid.ycellcenters
dis_lay0_depth = griddata(points_lay0, depth_lay0, (grid_x, grid_y), method = 'nearest')

dummy_botm[0,:,:] = dis_top - dis_lay0_depth

# adjust the first layer depth (upper aquifer bottom / aquitard top)
dummy_botm[0,:,:] = np.nanmin(dummy_botm[0,:,:]) + 0.4*(dis_top - np.nanmin(dummy_botm[0,:,:]))

# assign the second layer depth (aquitard bottom / lower aquifer top)
dis_lay1_botm = pd.read_csv('csv/DIS - layer 2 bottom.csv')
points_lay1 = dis_lay1_botm[['X', 'Y']].values
depth_lay1 = dis_lay1_botm['Relative depth'].values
grid_x = gwf.modelgrid.xcellcenters
grid_y = gwf.modelgrid.ycellcenters
dis_lay1_depth = griddata(points_lay1, depth_lay1, (grid_x, grid_y), method = 'nearest')

dummy_botm[1,:,:] = dummy_botm[0,:,:] - dis_lay1_depth

# update DIS_0.1 package with dummy subsurface layers
gwf.dis.botm = dummy_botm

#####################################################################################################################################

################################## IC package #######################################################################################

# assign the initial condition to all layers
ic_strt = np.zeros((nlay, nrow, ncol))
ic_strt[0, :, :] = dis_top - 10
ic_strt[1, :, :] = dis_top - 100
ic_strt[2, :, :] = dis_top - 200
ic_strt[3, :, :] = dis_top - 300

# define dummy initial conditions
ic = flopy.mf6.ModflowGwfic(gwf, strt = ic_strt)

#####################################################################################################################################

################################# NPF package #######################################################################################

# read shallow groundwater hydraulic conductivity from field data
k_1 = pd.read_csv('csv/NPF - layer 1 - field data - edit.csv')
k_1_points = k_1[['X (actual)', 'Y (actual)']].values
k_1_values = k_1['k'].values
k33_1_values = k_1['k33'].values

# read deep groundwater hydraulic conductivity from pumping tests
k_3 = pd.read_csv('csv/NPF - layer 3 - field data - edit.csv')
k_3_points = k_3[['X (actual)', 'Y (actual)']].values
k_3_values = k_3['k'].values

# interpolate using radial basis function from scipy.interpolate
rbf_npf_k1 = Rbf(k_1_points[:,0], k_1_points[:,1], k_1_values[:], function = "thin_plate")
grid_k_1_spline = rbf_npf_k1(grid_x, grid_y)

rbf_npf_k3 = Rbf(k_3_points[:,0], k_3_points[:,1], k_3_values[:], function = "thin_plate")
grid_k_3_spline = rbf_npf_k3(grid_x, grid_y)

rbf_npf_k33 = Rbf(k_1_points[:,0], k_1_points[:,1], k33_1_values[:], function = "thin_plate")
grid_k_33_spline = rbf_npf_k33(grid_x, grid_y)

k_factor = 1
aquitard_factor = 0.11

# create hydraulic conductivity array
flopy_npf_k = np.zeros((nlay, nrow, ncol))
flopy_npf_k[0,:,:] = k_factor*grid_k_1_spline
flopy_npf_k[1,:,:] = aquitard_factor*grid_k_1_spline
flopy_npf_k[2,:,:] = grid_k_3_spline*0.8
flopy_npf_k[3,:,:] = 0.0000001

flopy_npf_k33 = np.zeros((nlay, nrow, ncol))
flopy_npf_k33[0,:,:] = k_factor*grid_k_33_spline*10
flopy_npf_k33[1,:,:] = aquitard_factor*grid_k_33_spline*10
flopy_npf_k33[2,:,:] = grid_k_3_spline*1
flopy_npf_k33[3,:,:] = 0.0000001

# replace negative value in the interpolation to minimum value (applicable only to RBF interpolation)
flopy_npf_k[0,:,:][flopy_npf_k[0,:,:] < 0.0003504253562667259] = 0.0003504253562667259
flopy_npf_k[1,:,:][flopy_npf_k[1,:,:] < 3.504253562667259e-06] = 3.504253562667259e-06
flopy_npf_k[2,:,:][flopy_npf_k[2,:,:] < 0.0003504253562667259] = 0.0003504253562667259
flopy_npf_k33[0,:,:][flopy_npf_k33[0,:,:] < 5.987984615790642e-08] = 5.987984615790642e-08
flopy_npf_k33[1,:,:][flopy_npf_k33[1,:,:] < 5.987984615790642e-10] = 5.987984615790642e-10
flopy_npf_k33[2,:,:][flopy_npf_k33[2,:,:] < 5.987984615790642e-08] = 5.987984615790642e-08

# assign cell type
icelltype = np.zeros((nlay, nrow, ncol))
for i in basin_intersect.cellids:
    icelltype[0, i[0], i[1]] = 1
    icelltype[1, i[0], i[1]] = 0
    icelltype[2, i[0], i[1]] = 0
    icelltype[3, i[0], i[1]] = 0

# define node property flow (NPF) package, icelltype = 0 = confined
npf = flopy.mf6.ModflowGwfnpf(gwf, save_specific_discharge = True, icelltype = icelltype, k = flopy_npf_k, k33 = flopy_npf_k33,
                              rewet_record = ['wetfct', 1.0, 'iwetit', 1, 'ihdwet', 0], wetdry = 1.)

#####################################################################################################################################

################################## OBS utility #######################################################################################

# get shapefile boundary
obs = gpd.read_file('shp/observation point/data_layer1_2004/obs_lay1.shp')

# get observed groundwater level from csv file
obs_level = pd.read_csv('csv/OBS - 2004_lay1 - edit.csv')
obs_level_values = obs_level['Starting head'].values

# create rec.array of cell-id and observed groundwater level
obs_geometry = []
i = 0
for i in range(len(obs)):
    obs_geometry.append(ix.intersects(obs.iloc[i].geometry))
    i += 1
    
utlobs_input = []
j = 0
for j in range(len(obs)):
    utlobs_input.append([j, 'HEAD', (0, obs_geometry[j][0][0][0], obs_geometry[j][0][0][1])])
    j += 1

utlobs = flopy.mf6.ModflowUtlobs(gwf, continuous = utlobs_input)

#####################################################################################################################################

################################# RCH package #######################################################################################

# read wflow recharge output
rch = xr.open_dataset('netcdf/recharge_2004_forcalibration.nc')
rch = rch.to_array(name = 'recharge')
rch_nan = rch[1]
rch_mm = rch_nan.fillna(0.9)
rch_m = rch_mm/1000 #convert data in mm to m
rch_m = rch_m.astype(dtype = 'float64')

rch_list = []

i = 0
for i in range(len(basin_intersect)):
    rch_list.append([(0, basin_intersect[i][0][0], basin_intersect[i][0][1]), rch_m.values[0, basin_intersect[i][0][0], basin_intersect[i][0][1]]])
    i += 1

rch_input = {0: rch_list}

rch = flopy.mf6.ModflowGwfrch(gwf, stress_period_data = rch_input)

#####################################################################################################################################

################################# STO package #######################################################################################

# assign iconvert type
iconvert = np.zeros((nlay, nrow, ncol))
for i in basin_intersect.cellids:
    iconvert[0, i[0], i[1]] = 1
    iconvert[1, i[0], i[1]] = 0
    iconvert[2, i[0], i[1]] = 0
    iconvert[3, i[0], i[1]] = 0

#define dynamic hydraulic parameters: iconvert = 0 = confined
sto = flopy.mf6.ModflowGwfsto(gwf, save_flows = True, iconvert = iconvert, ss = 8.7e-3, sy = 0.2, steady_state = {0:True})

#####################################################################################################################################

################################# RIV package #######################################################################################

# get shapefile boundary & create the 'cell-line'
river = gpd.read_file('shp/river path/river_20220718.shp')
river_geometry = river.iloc[0].geometry
river_intersect = ix.intersects(river_geometry)

river_list = []
for i in river_intersect.cellids:
    river_list.append([(0, i[0], i[1]), dis_top[i[0], i[1]] - 21.0, 75, dis_top[i[0], i[1]] - 23.5])
river_spd = {0: river_list}

# define the RIV package
riv = flopy.mf6.ModflowGwfriv(gwf, stress_period_data = river_spd)

#####################################################################################################################################

################################# WEL package #######################################################################################

# get point location shapefile & its cellids for domestic abstraction
abs1_well = gpd.read_file('shp/abstraction point/domestic_new4.shp')

abs1_well_geometry = []
i = 0
for i in range(len(abs1_well)):
    abs1_well_geometry.append(ix.intersects(abs1_well.iloc[i].geometry))
    i += 1

# get point location shapefile & its cellids for industrial abstraction
abs2_well = gpd.read_file('shp/abstraction point/industry_point-river.shp')

abs2_well_geometry = []
i = 0
for i in range(len(abs2_well)):
    abs2_well_geometry.append(ix.intersects(abs2_well.iloc[i].geometry))
    i += 1

# assign the average domestic and industry abstraction
abs_well = flopy.mf6.ModflowGwfwel.stress_period_data.empty(gwf, maxbound = len(abs1_well_geometry) + len(abs2_well_geometry))

j = 0
for j in range(len(abs1_well)):
    abs_well[0][j] = ((0, abs1_well_geometry[j][0][0][0], abs1_well_geometry[j][0][0][1]), -1*263298.126/len(abs1_well_geometry))
    j += 1

j = len(abs1_well_geometry)
for j in range(len(abs1_well_geometry), len(abs1_well_geometry) + len(abs2_well_geometry), 1):
    abs_well[0][j] = ((2, abs2_well_geometry[j - len(abs1_well_geometry)][0][0][0],
                  abs2_well_geometry[j - len(abs1_well_geometry)][0][0][1]), -1*555851.5993/len(abs2_well_geometry))
    j += 1

well_spd = {0: abs_well[0]}

wel = flopy.mf6.ModflowGwfwel(gwf, stress_period_data = well_spd, save_flows = True)

#####################################################################################################################################

################################## OC package #######################################################################################

hname = '{}.hds'.format(modelname)
#cname = '{}.cbc'.format(modelname)
oc = flopy.mf6.ModflowGwfoc(gwf, head_filerecord = hname, saverecord = [('HEAD', 'ALL')])

#####################################################################################################################################

################################## MODEL RUN ########################################################################################

# write the model
sim.write_simulation()
sim.run_simulation()

#####################################################################################################################################

########################## Simulation post-processing ###############################################################################

head_bf = bf.HeadFile('steady-state_result/gwf_model.hds')
head = head_bf.get_data(totim = 1.0)
head[head == 1e+30] = np.nan

#cpth = os.path.join(model_ws, modelname + '.cbc')
#cobj = flopy.utils.CellBudgetFile(cpth, precision = head_bf.precision)
#spd = cobj.get_data(text = 'DATA-SPDIS')[0]

# get shapefile boundary
obs_geometry = []
i = 0
for i in range(len(obs)):
    obs_geometry.append(ix.intersects(obs.iloc[i].geometry))
    i += 1

obsvssim = []

j = 0
for j in range(len(obs)):
    obsvssim.append([j, obs_geometry[j][0][0][0], obs_geometry[j][0][0][1], -1*obs_level_values[j],
                     (dis_top[obs_geometry[j][0][0][0]][obs_geometry[j][0][0][1]] -
                     head[0][obs_geometry[j][0][0][0]][obs_geometry[j][0][0][1]]),
                     (dis_top[obs_geometry[j][0][0][0]][obs_geometry[j][0][0][1]] + obs_level_values[j]),
                     head[0][obs_geometry[j][0][0][0]][obs_geometry[j][0][0][1]]])
    j += 1

np.savetxt('obsvssim_spinup_1.csv', obsvssim, delimiter = ',')

# save the calibration result as csv
calibration_csv = []

i = 0
j = 0
for i in range(nrow):
    for j in range(ncol):
        calibration_csv.append([grid_x[0][j], grid_y[i][0], head[0][i][j]])
        j += 1            
    i += 1

np.savetxt('head_clip4_steady_newriver.csv', calibration_csv, delimiter = ',')

#####################################################################################################################################

# plot the mapview check
#fig, ax1 = plt.subplots(figsize = (15,10))
#topview = flopy.plot.PlotMapView(model = gwf, ax = ax1)
#linecollection = topview.plot_grid(linewidth = 0.1)
#surface_elevation = topview.plot_array(dis_top, alpha = 0.5)
#k = 0
#for k in range(len(abs_well)):
#    abstraction_plot = topview.plot_vector(vx = abs_well_geometry[k], vy = well[0][k][1])
#    k += 1
#cb_1 = fig.colorbar(surface_elevation, ax = ax1, shrink = 0.5, orientation = 'horizontal')
#cb_1.set_label('Surface elevation')
#head_levels = np.linspace(np.nanmin(head[0]), np.nanmax(head[0]), num = 20)
#head_plot = topview.contour_array(head[0], levels = head_levels, cmap = 'plasma')
#head_plot = topview.plot_array(head[0], cmap = 'YlGnBu')
#cbaxes = inset_axes(ax1, width = '30%', height = '3%', loc = 4)
#cb_2 = fig.colorbar(head_plot, ax = ax1, format = '%.0f', cax = cbaxes, orientation = 'horizontal')
#cb_2.set_label('Simulated heads')
#cb_2.mappable.set_clim(600, 1600)
#cbaxes.xaxis.set_ticks_position('top')
#quiver = topview.plot_specific_discharge(spd, istep = 20, jstep = 20)
#riv_plot = river.plot(ax = ax1, edgecolor = 'red', linewidth = 1)
#bc = topview.plot_ibound()
#actives = topview.plot_inactive(color_noflow = 'white')
#basin.plot(ax = ax1, facecolor = 'none', edgecolor = 'black', linewidth = 2)
#ax1.set_xlim(760000, 830000)
#ax1.set_xticklabels([])
#ax1.set_ylim(9195000, 9250000)
#ax1.set_yticklabels([])
#fig.show()

# plot the cross-section check
#fig = plt.figure(figsize = (15,10))
#ax2 = fig.add_subplot(1, 1, 1)
#x_section = flopy.plot.PlotCrossSection(model = gwf, ax = ax2, line = {'Row': 155})
#lines = x_section.plot_grid(linewidth = 0.05)
#hk_plot = x_section.plot_array(np.log(flopy_npf_k))
#bc = x_section.plot_ibound(head = head)
#ic_plot = x_section.plot_array(flopy_ic_strt_plotxs)
#ic_plot_xs_colorbar = fig.colorbar(ic_plot)
#ic_line = x_section.plot_surface(flopy_ic_strt_plotxs[0], masked_values = [999.], color = 'blue', lw = 2)
#head_plot_xs = x_section.plot_array(head, head = head, alpha = 0.5, cmap = 'RdBu')
#head_line = x_section.plot_surface(head[0], masked_values = [999.], color = 'blue', lw = 2)
#head_plot_xs_colorbar = fig.colorbar(head_plot_xs)
#ax2.set_ylim(300, 2500)
#fig.show()